#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

from collections                   import defaultdict
from cve_manager.knowledge_handler import KnowledgeHandler
from ax.ver                        import CompareSVersions

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Class that forms and stores a dict of issues that should not be presented

class ExcludedIssues(KnowledgeHandler):

	def __init__(self, conf_dir):

		super().__init__(conf_dir, 'excluded-issues.csv')

		self.excluded_issues = None
		self.ready = False

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Read excluded issues from a file specified at the initialization

	def Read(self):

		ok, msg = super().Check()
		if not ok:
			return None, msg

		excluded_issues = defaultdict(lambda: defaultdict(dict))
		warn = []
		first_line = True

		with open(self.file_path, 'r') as f:

			for line in f.readlines():

				line = line.strip()

				if not line:
					continue

				if first_line:
					first_line = False
					continue

				fields = [field.strip() for field in line.split(',')]
				if len(fields) < 4:
					warn.append(f'Wrong format of the "{self.file_path}" '
						f'file: {line}')
					continue

				# There could be multiple branch names, multiple vulnerability
				# ids and multiple package names separated with spaces
				branches  = [branch.replace('.', '') for branch in fields[0].split()]
				vul_ids   = fields[1].split()
				packages  = fields[2].split()
				fixed_ver = fields[3]

				for branch in branches:
					for vul_id in vul_ids:
						for package in packages:
							excluded_issues[branch][vul_id][package] = fixed_ver

		return excluded_issues, '; '.join(warn)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Check that given issue is excluded

	def IsExcluded(self, branch, vul_id, package_name, package_ver):

		msg = ''

		if not self.ready or self.excluded_issues == None:
			self.excluded_issues, msg = self.Read()
			if self.excluded_issues == None:
				return None, msg
			self.ready = True

		selection = {}

		for target in (branch, 'all'):
			ver = self.excluded_issues \
					.get(target, {})   \
					.get(vul_id, {})   \
					.get(package_name)
			if ver != None:
				selection[target] = ver

		if not selection:
			return False, msg

		# If there is an entry for a specific branch and an entry
		# for all branches
		if len(selection.values()) > 1:
			comp, _ = CompareSVersions(selection[branch], selection['all'])
			excluded_ver = selection[branch] \
				if comp <= 0 \
				else selection['all']
		else:
			excluded_ver = selection[branch] \
				if branch in selection \
				else selection['all']

		return CompareSVersions(excluded_ver, package_ver)[0] <= 0, msg
