#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import re
from collections import defaultdict
from ax.ver      import SVerToLVer, CompareLVersions, CompareSVerWithSRange

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

class Worker():

	def __init__ (self, vul_software, package_ver, saved_vul_versions, fix):

		self.vul_software = vul_software
		self.package_ver = package_ver
		self.saved_vul_versions = saved_vul_versions
		self.fix = fix

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Check if there is no generic ver in a list in a relation to a given one
	# (where ver is a list, for example, [1, 2, 3] for ver '1.2.3')

	def __GenericVer(self, ver, list_of_ver):

		ver_len = len(ver)

		for listed_ver in list_of_ver:
			listed_ver_len = len(listed_ver)
			# Ver with more stages of versioning can't be generic
			if listed_ver_len > ver_len:
				continue
			generic_in_the_list = True
			for i in range(0, listed_ver_len):
				if ver[i] != listed_ver[i]:
					generic_in_the_list = False
					break
			if generic_in_the_list:
				return False

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Forming a 'fix' record based on comparison

	def __Fix(self, comp, default_status='?'):

		# Check if a current fix is a 'definite' one (changelog, excluded)
		def IsDefiniteFix():
			return any(s in self.fix for s in ('c', 'e'))
		# - nested func

		if comp == None:
			if not IsDefiniteFix():
				return default_status
		elif comp > 0:
			return 'b'
		elif IsDefiniteFix():
			pass
		elif comp == 0:
			return 'V'
		else:
			return 'L'

		return ''

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Check if that is a comparison between numeric and symbolic versions

	def __NumVsSymb(self, symbolic):
		return symbolic in (1, 2)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Check if that is not a match and there is no generic ver to a given one

	def __NotMatchNotGeneric(self, comp, vul_lver):
		return comp != 0 and not self.__GenericVer(vul_lver,
			self.saved_vul_versions[self.vul_software])

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Forming a fix record for a single vulnerable version

	def SingleVerFix(self, vul_ver):

		if not vul_ver:
			return ''

		# If vulnerable version is not stated
		if vul_ver == '-':
			return self.__Fix(None, 'V')

		package_lver = SVerToLVer(self.package_ver)
		vul_lver = SVerToLVer(vul_ver)
		comp, symbolic = CompareLVersions(package_lver, vul_lver)

		# If one version is symbolic and another one is not
		if abs(comp) <= 1 and self.__NumVsSymb(symbolic):
			return self.__Fix(None)

		# If versions not match and none of them is generic to another one
		if self.__NotMatchNotGeneric(comp, vul_lver):
			return ''

		self.saved_vul_versions[self.vul_software].append(vul_lver)

		return self.__Fix(comp)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Forming a fix record for a list of versions (in form of a string)

	def ListVerFix(self, vul_ver):

		if not vul_lver_list:
			return ''

		comp_list = []
		package_lver = SVerToLVer(self.package_ver)
		vul_lver_list = [SVerToLVer(v) for v in vul_ver.split(', ')]

		for i, vul_lver in enumerate(vul_lver_list):
			# If vulnerable version is not stated
			if not vul_lver or vul_lver[0] == '-':
				continue
			comp, symbolic = CompareLVersions(package_lver, vul_lver)
			# If one version is symbolic and another one is not
			if (abs(comp) <= 1 and self.__NumVsSymb(symbolic)) or \
					self.__NotMatchNotGeneric(comp, vul_lver):
				continue
			self.saved_vul_versions[self.vul_software].append(vul_lver)
			comp_list.append(comp)

		if not comp_list:
			return ''

		comp = 0
		if all(c > 0 for c in comp_list):
			comp = 1
		elif all(c < 0 for c in comp_list):
			comp = -1

		return self.__Fix(comp)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Forming a fix record for a range of versions

	def RangeVerFix(self, vul_ver):

		comp, symbolic_left, symbolic_right = \
			CompareSVerWithSRange(self.package_ver, vul_ver)

		if comp == None:
			return ''

		# If there was a symbolic comparison that was triggered
		# by a range limit only
		if any(s == 2 for s in (symbolic_left, symbolic_right)):
			return self.__Fix(None)

		return self.__Fix(comp)

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Forming a 'fix' record and a 'vulnerable version' record

def GetFixRecord(package_ver, vul_ver, vul_software='',
		saved_vul_versions=defaultdict(list), fix=''):

	# Worker obj will do the job of producing a 'fix' record,
	# the only problem here is to pick a right method to call
	w = Worker(vul_software, package_ver, saved_vul_versions, fix)

	if not vul_ver:
		return '?'

	if vul_ver[0] in '([':
		return w.RangeVerFix(vul_ver)
	elif ',' in vul_ver:
		return w.ListVerFix(vul_ver)

	return w.SingleVerFix(vul_ver)
