#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

from os                   import cpu_count
from collections          import defaultdict
from mysql.connector      import Error as MySQLError
from ax.ver               import IsDateVer
from cve_manager.mediator import ProtoMediator
from cve_manager.common   import SplitData
from cve_manager.conf     import LOCAL_SYS
from cve_manager.db       import DB
from cve_manager.url      import ParseAndJoinURL

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Interface between module that perform issues detection and a DB

class Mediator(ProtoMediator):

	def __init__(self, mysql_config, printer):

		super().__init__(mysql_config, printer)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Create issues table for a giving branch

	def RecreateIssuesTable(self, branch, data_source):

		table = f'{branch}_{data_source}_issues'
		vul_rec_table = f'{data_source}_vul_import'

		self.printer.LineBegin(f'Creating {table} table')

		# Connecting to the CVE database
		db = super().NewDB()
		if not db:
			return False

		# Don't recreate the table if there is no corresponding table
		# and don't do detection either
		if not db.TableExists(vul_rec_table):
			self.printer.Err('There is no table with vulnerabilities records '
				'from ' + data_source.upper())
			return False

		# Forming the table description
		columns = [
			'vul_year YEAR',
			'vul_num VARCHAR(8)',
			'package_id INT UNSIGNED',
			'vul_ver TEXT',
			'fix VARCHAR(3)',
			f'CONSTRAINT pk_{table} '
				'PRIMARY KEY (vul_year, vul_num, package_id, fix)',
			f'CONSTRAINT {table}_fk_{branch} '
				f'FOREIGN KEY (package_id) REFERENCES {branch}_src (id)',
			f'CONSTRAINT {table}_fk_{vul_rec_table} '
				'FOREIGN KEY (vul_year, vul_num) '
					f'REFERENCES {vul_rec_table} (vul_year, vul_num)',
			]

		if not db.RecreateTable(table, columns):
			return False

		self.printer.Success()

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Check that table for issues of a specified branch and of a specified data
	# source exists

	def CheckIssuesTable(self, branch, data_source):

		db = super().NewDB()
		if not db:
			return False

		issues_table = f'{branch}_{data_source}_issues'

		if not db.TableExists(issues_table):
			self.printer.Err(f'Table "{issues_table}" doesn\'t exist, pass the '
				'"--prepare" flag to recreate *_issues tables for all branches '
				'that being processed')
			return False

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get package descriptions of a specified branch

	def GetPackages(self, branch, data_source, packages):

		msg = 'Getting a dict of package descriptions of the ' \
			f'{"local system" if branch == LOCAL_SYS else f"{branch} branch"} ' \
			f'for the {data_source.upper()} data source'
		self.printer.LineBegin(msg)

		db = super().NewDB()
		if not db:
			return {}

		packages = defaultdict(list)
		condition = f"{data_source}_name <> '' " \
			f"AND {data_source}_name is not NULL"

		packages_cursor = db.QueryPackages(branch, packages, [],
			data_source, custom_condition=condition, mapping_name=True,
			url=True, fixes=True)

		if packages_cursor == None:
			return {}

		for unique_id, name, ver, rel, map_names, url, fixes in packages_cursor:
			is_date_ver = IsDateVer(ver)
			url = ParseAndJoinURL(url)
			package_entry = (
				unique_id, name, ver, is_date_ver, rel, url, set(fixes.split())
				)
			for map_name in map_names.split():
				packages[map_name].append(package_entry)

		if not packages:
			msg = 'Empty dict'
			if data_source == FSTEC_DATA_SRC:
				self.printer.Warn(msg)
			else:
				self.printer.Err(msg)
		else:
			msg = f'Got entries for {str(len(packages))} product names'
			self.printer.Success(msg)

		return packages

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get a list of lists of vulnerabilities (each list for a separate process)

	def GetVulnerabilities(self, data_source):

		self.printer.LineBegin(f'Getting a list of {data_source.upper()} '
			'vulnerabilities')

		db = super().NewDB()
		if not db:
			return []

		# Getting the list
		table = f'{data_source}_vul_import'
		cols  = ['vul_year', 'vul_num', 'software_list']
		vul_cursor = db.Select(cols, table)
		if not vul_cursor:
			if vul_cursor != None:
				self.printer.Err('Empty list')
			return []

		# Splitting the data according to the number of cores
		vul = SplitData(vul_cursor, cpu_count())

		entries_n = sum(len(v) for v in vul)
		msg = f'{entries_n} entries'

		if entries_n > 0:
			self.printer.Success(msg)
		else:
			self.printer.Warn(msg)

		return vul

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Send issues of a specified branch found for a specified data source (this
	# method is different since it may be executed in different threads
	# simultaneously)

	def SendIssues(self, branch, data_source, data):

		db = super().NewDB()
		if not db:
			return 'Can\'t connect to the DB'

		query = f'INSERT INTO {branch}_{data_source}_issues ' \
			'(vul_year, vul_num, package_id, fix, vul_ver) ' \
			'VALUES (%s, %s, %s, %s, %s)'

		if db.Cursor(query, data=data) == None:
			return 'Can\'t send issues'

		return ''
